@@ -346,7 +346,13 @@
  * The number of bytes that are needed for a redirector stub.
  * This includes any extra bytes that are needed for alignment.
  */
-#define	jit_redirector_size		32
+#define	jit_redirector_size		24
+
+/*
+ * The number of bytes that are needed for a indirector stub.
+ * This includes any extra bytes that are needed for alignment.
+ */
+#define	jit_indirector_size		8
 
 /*
  * We should pad unused code space with NOP's.
