@@ -419,6 +419,22 @@ void jit_dump_insn(FILE *stream, jit_function_t func, jit_insn_t insn)
 				(long)(jit_insn_get_label(insn)));
 		return;
 	}
+	else if((flags & JIT_OPCODE_IS_JUMP_TABLE) != 0)
+	{
+		jit_label_t *labels;
+		jit_nint num_labels, label;
+		labels = (jit_label_t *)jit_value_get_nint_constant(jit_insn_get_value1(insn));
+		num_labels = jit_value_get_nint_constant(jit_insn_get_value2(insn));
+		fprintf(stream, "jump_table ");
+		dump_value(stream, func, jit_insn_get_dest(insn), flags & JIT_OPCODE_DEST_MASK);
+		printf(" : {");
+		for(label = 0; label < num_labels; label++)
+		{
+			printf(" .L%ld", (long) labels[label]);
+		}
+		printf(" }");
+		return;
+	}
 
 	/* Output the destination information */
 	if((flags & JIT_OPCODE_DEST_MASK) != JIT_OPCODE_DEST_EMPTY &&
@@ -629,6 +645,17 @@ static void dump_interp_code(FILE *stream, void **pc, void **end)
 							(long)(jit_nint)(pc[1]), (long)(jit_nint)(pc[2]));
 					pc += 3;
 				}
+				else if((info->flags & JIT_OPCODE_IS_JUMP_TABLE) != 0)
+				{
+					jit_nint label, num_labels;
+					num_labels = (jit_nint)pc[0];
+					for(label = 1; label <= num_labels; label++)
+					{
+						fprintf(stream,	" %lX",
+							(long)(jit_nint)pc[label]);
+					}
+					pc += 1 + num_labels;
+				}
 			}
 			break;
 		}
@@ -641,14 +668,14 @@ static void dump_interp_code(FILE *stream, void **pc, void **end)
 #else /* !JIT_BACKEND_INTERP */
 
 /*
- * Dump the native object code representation of a function.
- * Can only dump to stdout or stderr at the moment.
+ * Dump the native object code representation of a function to stream.
  */
 static void dump_object_code(FILE *stream, void *start, void *end)
 {
 	char cmdline[BUFSIZ];
 	unsigned char *pc = (unsigned char *)start;
 	FILE *file;
+	int ch;
 
 #if JIT_WIN32_PLATFORM
 	/*
@@ -693,9 +720,18 @@ static void dump_object_code(FILE *stream, void *start, void *end)
 	fclose(file);
 	sprintf(cmdline, "as %s -o %s", s_path, o_path);
 	system(cmdline);
-	sprintf(cmdline, "objdump --adjust-vma=%ld -d %s%s",
-			(long)(jit_nint)start, o_path, (stream == stderr ? " 1>&2" : ""));
+	sprintf(cmdline, "objdump --adjust-vma=%ld -d %s > %s",
+			(long)(jit_nint)start, o_path, s_path);
 	system(cmdline);
+	file = fopen(s_path, "r");
+	if(file)
+	{
+		while((ch = getc(file)) != EOF)
+		{
+			putc(ch, stream);
+		}
+		fclose(file);
+	}
 	unlink(s_path);
 	unlink(o_path);
 	putc('\n', stream);
@@ -706,7 +742,7 @@ static void dump_object_code(FILE *stream, void *start, void *end)
 
 /*@
  * @deftypefun void jit_dump_function ({FILE *} stream, jit_function_t func, {const char *} name)
- * Dump the three-address instructions within a function to a stdio stream.
+ * Dump the three-address instructions within a function to a stream.
  * The @code{name} is attached to the output as a friendly label, but
  * has no other significance.
  *
@@ -856,4 +892,5 @@ void jit_dump_function(FILE *stream, jit_function_t func, const char *name)
 
 	/* Output the function footer */
 	fprintf(stream, "end\n\n");
+	fflush(stream);
 }
