@@ -24,6 +24,9 @@
 #include "jit-reg-alloc.h"
 #include "jit-apply-func.h"
 #include "jit-setjmp.h"
+#ifdef _JIT_COMPILE_DEBUG
+#include <stdio.h>
+#endif
 
 /*@
  * @deftypefun jit_function_t jit_function_create (jit_context_t context, jit_type_t signature)
@@ -46,6 +49,12 @@
 jit_function_t jit_function_create(jit_context_t context, jit_type_t signature)
 {
 	jit_function_t func;
+#if defined(jit_redirector_size)
+	jit_on_demand_driver_func on_demand_driver;
+#endif
+#if defined(jit_redirector_size) || defined(jit_indirector_size)
+	jit_cache_t cache;
+#endif
 
 	/* Allocate memory for the function and clear it */
 	func = jit_cnew(struct _jit_function);
@@ -54,19 +63,68 @@ jit_function_t jit_function_create(jit_context_t context, jit_type_t signature)
 		return 0;
 	}
 
+#if defined(jit_redirector_size) || defined(jit_indirector_size)
+	/* TODO: if the function is destroyed the redirector and indirector memory
+	   is leaked */
+
+	/* We need the cache lock while we are allocating redirector and indirector */
+	jit_mutex_lock(&(context->cache_lock));
+
+	/* Get the method cache */
+	cache = _jit_context_get_cache(context);
+	if(!cache)
+	{
+		jit_mutex_unlock(&(context->cache_lock));
+		jit_free(func);
+		return 0;
+	}
+
+# if defined(jit_redirector_size)
+	/* Allocate redirector buffer */
+	func->redirector = _jit_cache_alloc_no_method(cache, jit_redirector_size, 1);
+	if(!func->redirector)
+	{
+		jit_mutex_unlock(&(context->cache_lock));
+		jit_free(func);
+		return 0;
+	}
+# endif
+# if defined(jit_indirector_size)
+	/* Allocate indirector buffer */
+	func->indirector = _jit_cache_alloc_no_method(cache, jit_indirector_size, 1);
+	if(!func->indirector)
+	{
+		jit_mutex_unlock(&(context->cache_lock));
+		jit_free(func);
+		return 0;
+	}
+# endif
+
+	jit_mutex_unlock(&(context->cache_lock));
+#endif
+
 	/* Initialize the function block */
 	func->context = context;
 	func->signature = jit_type_copy(signature);
 
-#if defined(jit_redirector_size) && !defined(JIT_BACKEND_INTERP)
+#if defined(jit_redirector_size)
 	/* If we aren't using interpretation, then point the function's
 	   initial entry point at the redirector, which in turn will
 	   invoke the on-demand compiler */
+	on_demand_driver = context->on_demand_driver;
+	if(!on_demand_driver)
+	{
+		on_demand_driver = _jit_function_compile_on_demand;
+	}
 	func->entry_point = _jit_create_redirector
-		(func->redirector, (void *)_jit_function_compile_on_demand,
+		(func->redirector, (void *) on_demand_driver,
 		 func, jit_type_get_abi(signature));
-	func->closure_entry = func->entry_point;
+	jit_flush_exec(func->redirector, jit_redirector_size);
 #endif
+# if defined(jit_indirector_size)
+	_jit_create_indirector(func->indirector, (void**) &(func->entry_point));
+	jit_flush_exec(func->indirector, jit_indirector_size);
+# endif
 
 	/* Add the function to the context list */
 	func->next = 0;
@@ -130,6 +188,11 @@ int _jit_function_ensure_builder(jit_function_t func)
 		return 0;
 	}
 
+	/* Cache the value of the JIT_OPTION_POSITION_INDEPENDENT option */
+	func->builder->position_independent
+		= jit_context_get_meta_numeric(
+			func->context, JIT_OPTION_POSITION_INDEPENDENT);
+
 	/* Initialize the function builder */
 	jit_memory_pool_init(&(func->builder->value_pool), struct _jit_value);
 	jit_memory_pool_init(&(func->builder->insn_pool), struct _jit_insn);
@@ -440,10 +503,31 @@ static void compile_block(jit_gencode_t gen, jit_function_t func,
 	jit_insn_iter_t iter;
 	jit_insn_t insn;
 
+#ifdef _JIT_COMPILE_DEBUG
+	printf("Block #%d: %d\n", func->builder->block_count++, block->label);
+#endif
+
 	/* Iterate over all blocks in the function */
 	jit_insn_iter_init(&iter, block);
 	while((insn = jit_insn_iter_next(&iter)) != 0)
 	{
+#ifdef _JIT_COMPILE_DEBUG
+		unsigned char *p1, *p2;
+		p1 = gen->posn.ptr;
+		printf("Insn: %5d, Opcode: 0x%04x\n", func->builder->insn_count++, insn->opcode);
+		printf("Start of binary code: 0x%08x\n", p1);
+#endif
+
+		/* The cache is full. */
+		if(!jit_cache_check_for_n(&gen->posn, 1))
+		{
+#ifdef _JIT_COMPILE_DEBUG
+			printf("No space left in the code cache.\n\n");
+			fflush(stdout);
+#endif
+			return;
+		}
+
 		switch(insn->opcode)
 		{
 			case JIT_OP_NOP:		break;		/* Ignore NOP's */
@@ -458,6 +542,7 @@ static void compile_block(jit_gencode_t gen, jit_function_t func,
 			}
 			break;
 
+#ifndef JIT_BACKEND_INTERP
 			case JIT_OP_INCOMING_REG:
 			{
 				/* Assign a register to an incoming value */
@@ -466,6 +551,7 @@ static void compile_block(jit_gencode_t gen, jit_function_t func,
 					 insn->value1);
 			}
 			break;
+#endif
 
 			case JIT_OP_INCOMING_FRAME_POSN:
 			{
@@ -477,7 +563,7 @@ static void compile_block(jit_gencode_t gen, jit_function_t func,
 				if(insn->value1->has_global_register)
 				{
 					insn->value1->in_global_register = 1;
-					_jit_gen_load_global(gen, insn->value1);
+					_jit_gen_load_global(gen, insn->value1->global_reg, insn->value1);
 				}
 				else
 				{
@@ -486,6 +572,7 @@ static void compile_block(jit_gencode_t gen, jit_function_t func,
 			}
 			break;
 
+#ifndef JIT_BACKEND_INTERP
 			case JIT_OP_OUTGOING_REG:
 			{
 				/* Copy a value into an outgoing register */
@@ -494,6 +581,7 @@ static void compile_block(jit_gencode_t gen, jit_function_t func,
 					 insn->value1);
 			}
 			break;
+#endif
 
 			case JIT_OP_OUTGOING_FRAME_POSN:
 			{
@@ -508,6 +596,7 @@ static void compile_block(jit_gencode_t gen, jit_function_t func,
 			}
 			break;
 
+#ifndef JIT_BACKEND_INTERP
 			case JIT_OP_RETURN_REG:
 			{
 				/* Assign a register to a return value */
@@ -517,6 +606,7 @@ static void compile_block(jit_gencode_t gen, jit_function_t func,
 				_jit_gen_insn(gen, func, block, insn);
 			}
 			break;
+#endif
 
 			case JIT_OP_MARK_OFFSET:
 			{
@@ -535,43 +625,103 @@ static void compile_block(jit_gencode_t gen, jit_function_t func,
 			}
 			break;
 		}
+
+#ifdef _JIT_COMPILE_DEBUG
+		p2 = gen->posn.ptr;
+		printf("Length of binary code: %d\n\n", p2 - p1);
+		fflush(stdout);
+#endif
 	}
 }
 
 /*
- * Information that is stored for an exception region in the cache.
+ * Reset value on restart.
  */
-typedef struct jit_cache_eh *jit_cache_eh_t;
-struct jit_cache_eh
+static void
+reset_value(jit_value_t value)
 {
-	jit_label_t		handler_label;
-	unsigned char  *handler;
-	jit_cache_eh_t	previous;
-};
+	value->reg = -1;
+	value->in_register = 0;
+	value->in_global_register = 0;
+	value->in_frame = 0;
+}
 
-/*@
- * @deftypefun int jit_function_compile (jit_function_t func)
- * Compile a function to its executable form.  If the function was
- * already compiled, then do nothing.  Returns zero on error.
- *
- * If an error occurs, you can use @code{jit_function_abandon} to
- * completely destroy the function.  Once the function has been compiled
- * successfully, it can no longer be abandoned.
- *
- * Sometimes you may wish to recompile a function, to apply greater
- * levels of optimization the second time around.  You must call
- * @code{jit_function_set_recompilable} before you compile the function
- * the first time.  On the second time around, build the function's
- * instructions again, and call @code{jit_function_compile}
- * a second time.
- * @end deftypefun
-@*/
-int jit_function_compile(jit_function_t func)
+/*
+ * Clean up the compilation state on restart.
+ */
+static void
+cleanup_on_restart(jit_gencode_t gen, jit_function_t func)
+{
+	jit_block_t block;
+	jit_insn_iter_t iter;
+	jit_insn_t insn;
+
+#ifdef _JIT_COMPILE_DEBUG
+	printf("\n*** Restart compilation ***\n\n");
+#endif
+
+	block = 0;
+	while((block = jit_block_next(func, block)) != 0)
+	{
+		/* Clear the block addresses and fixup lists */
+		block->address = 0;
+		block->fixup_list = 0;
+		block->fixup_absolute_list = 0;
+
+		/* Reset values referred to by block instructions */
+		jit_insn_iter_init(&iter, block);
+		while((insn = jit_insn_iter_next(&iter)) != 0)
+		{
+			if(insn->dest && (insn->flags & JIT_INSN_DEST_OTHER_FLAGS) == 0)
+			{
+				reset_value(insn->dest);
+			}
+			if(insn->value1 && (insn->flags & JIT_INSN_VALUE1_OTHER_FLAGS) == 0)
+			{
+				reset_value(insn->value1);
+			}
+			if(insn->value2 && (insn->flags & JIT_INSN_VALUE2_OTHER_FLAGS) == 0)
+			{
+				reset_value(insn->value2);
+			}
+		}
+	}
+
+	/* Reset values referred to by builder */
+	if(func->builder->setjmp_value)
+	{
+		reset_value(func->builder->setjmp_value);
+	}
+	if(func->builder->parent_frame)
+	{
+		reset_value(func->builder->parent_frame);
+	}
+
+	/* Reset the "touched" registers mask. The first time compilation
+	   might have followed wrong code paths and thus allocated wrong
+	   registers. */
+	if(func->builder->has_tail_call)
+	{
+		/* For functions with tail calls _jit_regs_alloc_global()
+		   does not allocate any global registers. The "permanent"
+		   mask has all global registers set to prevent their use. */
+		gen->touched = jit_regused_init;
+	}
+	else
+	{
+		gen->touched = gen->permanent;
+	}
+}
+
+/*
+ * Compile a function and return its entry point.
+ */
+static int
+compile(jit_function_t func, void **entry_point)
 {
 	struct jit_gencode gen;
 	jit_cache_t cache;
 	void *start;
-	void *recompilable_start = 0;
 	void *end;
 	jit_block_t block;
 	int result;
@@ -579,22 +729,6 @@ int jit_function_compile(jit_function_t func)
 	int have_prolog;
 #endif
 
-	/* Bail out if we have nothing to do */
-	if(!func)
-	{
-		return 0;
-	}
-	if(func->is_compiled && !(func->builder))
-	{
-		/* The function is already compiled, and we don't need to recompile */
-		return 1;
-	}
-	if(!(func->builder))
-	{
-		/* We don't have anything to compile at all */
-		return 0;
-	}
-
 	/* We need the cache lock while we are compiling the function */
 	jit_mutex_lock(&(func->context->cache_lock));
 
@@ -613,12 +747,20 @@ int jit_function_compile(jit_function_t func)
 	_jit_function_compute_liveness(func);
 
 	/* Allocate global registers to variables within the function */
+#ifndef JIT_BACKEND_INTERP
 	_jit_regs_alloc_global(&gen, func);
+#endif
 
 	/* We may need to perform output twice, if the first attempt fails
 	   due to a lack of space in the current method cache page */
 	do
 	{
+#ifdef _JIT_COMPILE_DEBUG
+		printf("\n*** Start compilation ***\n\n");
+		func->builder->block_count = 0;
+		func->builder->insn_count = 0;
+#endif
+
 		/* Start function output to the cache */
 		start = _jit_cache_start_method
 			(cache, &(gen.posn), JIT_FUNCTION_ALIGNMENT, func);
@@ -651,7 +793,9 @@ int jit_function_compile(jit_function_t func)
 #endif
 
 		/* Clear the register assignments for the first block */
+#ifndef JIT_BACKEND_INTERP
 		_jit_regs_init_for_block(&gen);
+#endif
 
 		/* Generate code for the blocks in the function */
 		block = 0;
@@ -668,15 +812,19 @@ int jit_function_compile(jit_function_t func)
 	
 			/* Generate the block's code */
 			compile_block(&gen, func, block);
-	
+
 			/* Spill all live register values back to their frame positions */
+#ifndef JIT_BACKEND_INTERP
 			_jit_regs_spill_all(&gen);
+#endif
 
 			/* Notify the back end that the block is finished */
 			_jit_gen_end_block(&gen, block);
 	
 			/* Clear the local register assignments, ready for the next block */
+#ifndef JIT_BACKEND_INTERP
 			_jit_regs_init_for_block(&gen);
+#endif
 		}
 
 		/* Output the function epilog.  All return paths will jump to here */
@@ -691,29 +839,26 @@ int jit_function_compile(jit_function_t func)
 		}
 #endif
 
+#if !defined(jit_redirector_size) || !defined(jit_indirector_size)
 		/* If the function is recompilable, then we need an extra entry
 		   point to properly redirect previous references to the function */
-		if(func->is_recompilable)
+		if(func->is_recompilable && !func->indirector)
 		{
-			recompilable_start = _jit_gen_redirector(&gen, func);
+			/* TODO: use _jit_create_indirector() instead of
+			   _jit_gen_redirector() as both do the same. */
+			func->indirector = _jit_gen_redirector(&gen, func);
 		}
+#endif
 
 		/* End the function's output process */
 		result = _jit_cache_end_method(&(gen.posn));
 
-		/* If we need to restart on a different cache page, then clear
-		   the block addresses and fixup lists */
+		/* If we need to restart on a different cache page, then clean up
+		   the compilation state  */
 		if(result == JIT_CACHE_END_RESTART)
 		{
-			block = 0;
-			while((block = jit_block_next(func, block)) != 0)
-			{
-				block->address = 0;
-				block->fixup_list = 0;
-				block->fixup_absolute_list = 0;
-			}
+			cleanup_on_restart(&gen, func);
 		}
-
 	}
 	while(result == JIT_CACHE_END_RESTART);
 
@@ -745,26 +890,147 @@ int jit_function_compile(jit_function_t func)
 		func->no_return = 1;
 	}
 
-	/* Record the entry point */
-	func->entry_point = start;
-	if(recompilable_start)
-	{
-		func->closure_entry = recompilable_start;
-	}
-	else
-	{
-		func->closure_entry = start;
-	}
-	func->is_compiled = 1;
-
 	/* Free the builder structure, which we no longer require */
 	_jit_function_free_builder(func);
 
 	/* The function has been compiled successfully */
 	jit_mutex_unlock(&(func->context->cache_lock));
+
+	/* Record the entry point */
+	if(entry_point)
+	{
+		*entry_point = start;
+	}
+
 	return 1;
 }
 
+/*
+ * Information that is stored for an exception region in the cache.
+ */
+typedef struct jit_cache_eh *jit_cache_eh_t;
+struct jit_cache_eh
+{
+	jit_label_t		handler_label;
+	unsigned char  *handler;
+	jit_cache_eh_t	previous;
+};
+
+/*@
+ * @deftypefun int jit_function_compile (jit_function_t func)
+ * Compile a function to its executable form.  If the function was
+ * already compiled, then do nothing.  Returns zero on error.
+ *
+ * If an error occurs, you can use @code{jit_function_abandon} to
+ * completely destroy the function.  Once the function has been compiled
+ * successfully, it can no longer be abandoned.
+ *
+ * Sometimes you may wish to recompile a function, to apply greater
+ * levels of optimization the second time around.  You must call
+ * @code{jit_function_set_recompilable} before you compile the function
+ * the first time.  On the second time around, build the function's
+ * instructions again, and call @code{jit_function_compile}
+ * a second time.
+ * @end deftypefun
+@*/
+int jit_function_compile(jit_function_t func)
+{
+	int result;
+	void *entry_point;
+
+	/* Bail out if we have nothing to do */
+	if(!func)
+	{
+		return 0;
+	}
+	if(func->is_compiled && !(func->builder))
+	{
+		/* The function is already compiled, and we don't need to recompile */
+		return 1;
+	}
+	if(!(func->builder))
+	{
+		/* We don't have anything to compile at all */
+		return 0;
+	}
+
+	/* Compile and record the entry point. */
+	result = compile(func, &entry_point);
+	if(result)
+	{
+		func->entry_point = entry_point;
+		func->is_compiled = 1;
+	}
+
+	return result;
+}
+
+/*@
+ * @deftypefun int jit_function_compile_entry (jit_function_t func, void **entry_point)
+ * Compile a function to its executable form but do not make it
+ * available for invocation yet. It may be made available later
+ * with @code{jit_function_setup_entry}.
+ * @end deftypefun
+@*/
+int
+jit_function_compile_entry(jit_function_t func, void **entry_point)
+{
+	/* Init entry_point */
+	if(entry_point)
+	{
+		*entry_point = 0;
+	}
+	else
+	{
+		return 0;
+	}
+
+	/* Bail out if we have nothing to do */
+	if(!func)
+	{
+		return 0;
+	}
+	if(func->is_compiled && !(func->builder))
+	{
+		/* The function is already compiled, and we don't need to recompile */
+		*entry_point = func->entry_point;
+		return 1;
+	}
+	if(!(func->builder))
+	{
+		/* We don't have anything to compile at all */
+		return 0;
+	}
+
+	/* Compile and return the entry point. */
+	return compile(func, entry_point);
+}
+
+/*@
+ * @deftypefun int jit_function_setup_entry (jit_function_t func, void *entry_point)
+ * Make a function compiled with @code{jit_function_compile_entry}
+ * available for invocation and free the resources used for
+ * compilation. If @code{entry_point} is null then it only
+ * frees the resources.
+ * @end deftypefun
+@*/
+void
+jit_function_setup_entry(jit_function_t func, void *entry_point)
+{
+	/* Bail out if we have nothing to do */
+	if(!func)
+	{
+		return;
+	}
+	/* Record the entry point */
+	if(entry_point)
+	{
+		func->entry_point = entry_point;
+		func->is_compiled = 1;
+	}
+	_jit_function_free_builder(func);
+}
+
 /*@
  * @deftypefun int jit_function_recompile (jit_function_t func)
  * Force @code{func} to be recompiled, by calling its on-demand
@@ -937,7 +1203,11 @@ void *jit_function_to_closure(jit_function_t func)
 							  function_closure, (void *)func);
 #else
 	/* On native platforms, use the closure entry point */
-	return func->closure_entry;
+	if(func->indirector && (!func->is_compiled || func->is_recompilable))
+	{
+		return func->indirector;
+	}
+	return func->entry_point;
 #endif
 }
 
@@ -988,6 +1258,7 @@ jit_function_t jit_function_from_pc
 	/* Convert the cookie into a handler address */
 	if(handler)
 	{
+#if 0
 		if(cookie)
 		{
 			*handler = ((jit_cache_eh_t)cookie)->handler;
@@ -996,6 +1267,9 @@ jit_function_t jit_function_from_pc
 		{
 			*handler = 0;
 		}
+#else
+		*handler = cookie;
+#endif
 	}
 	return func;
 }
@@ -1020,14 +1294,43 @@ void *jit_function_to_vtable_pointer(jit_function_t func)
 	return func;
 #else
 	/* On native platforms, the closure entry point is the vtable pointer */
-	if(func)
-	{
-		return func->closure_entry;
-	}
-	else
+	if(!func)
 	{
 		return 0;
 	}
+	if(func->indirector && (!func->is_compiled || func->is_recompilable))
+	{
+		return func->indirector;
+	}
+	return func->entry_point;
+#endif
+}
+
+/*@
+ * @deftypefun jit_function_t jit_function_from_vtable_pointer (jit_context_t context, {void *} vtable_pointer)
+ * Convert a vtable_pointer back into a function.  Returns NULL if the
+ * vtable_pointer does not correspond to a function in the specified context.
+ * @end deftypefun
+@*/
+jit_function_t jit_function_from_vtable_pointer(jit_context_t context, void *vtable_pointer)
+{
+#ifdef JIT_BACKEND_INTERP
+	/* In the interpreted version, the function pointer is used in vtables */
+	jit_function_t func = (jit_function_t)vtable_pointer;
+
+	if(func && func->context == context)
+	{
+		return func;
+	}
+	return 0;
+#else
+	void *cookie;
+	if(!context || !(context->cache))
+	{
+		return 0;
+	}
+	return (jit_function_t)_jit_cache_get_method
+		(context->cache, vtable_pointer, &cookie);
 #endif
 }
 
@@ -1077,10 +1380,28 @@ void *jit_function_to_vtable_pointer(jit_function_t func)
  * just after you create it with @code{jit_function_create}.
  * @end deftypefun
 @*/
-void jit_function_set_on_demand_compiler
-		(jit_function_t func, jit_on_demand_func on_demand)
+void
+jit_function_set_on_demand_compiler(jit_function_t func, jit_on_demand_func on_demand)
 {
-	func->on_demand = on_demand;
+	if(func)
+	{
+		func->on_demand = on_demand;
+	}
+}
+
+/*@
+ * @deftypefun jit_on_demand_func jit_function_get_on_demand_compiler (jit_function_t func)
+ * Returns function's on-demand compiler.
+ * @end deftypefun
+@*/
+jit_on_demand_func
+jit_function_get_on_demand_compiler(jit_function_t func)
+{
+	if(func)
+	{
+		return func->on_demand;
+	}
+	return 0;
 }
 
 void *_jit_function_compile_on_demand(jit_function_t func)
@@ -1186,6 +1507,9 @@ int jit_function_apply_vararg
 	(jit_function_t func, jit_type_t signature, void **args, void *return_area)
 {
 	struct jit_backtrace call_trace;
+#if defined(jit_redirector_size)
+	jit_on_demand_driver_func on_demand_driver;
+#endif
 	void *entry;
 	jit_jmp_buf jbuf;
 
@@ -1220,7 +1544,16 @@ int jit_function_apply_vararg
 	}
 	else
 	{
+#if defined(jit_redirector_size)
+		on_demand_driver = func->context->on_demand_driver;
+		if(!on_demand_driver)
+		{
+			on_demand_driver = _jit_function_compile_on_demand;
+		}
+		entry = (*on_demand_driver)(func);
+#else
 		entry = _jit_function_compile_on_demand(func);
+#endif
 	}
 
 	/* Get the default signature if necessary */
