@@ -27,6 +27,11 @@
 extern	"C" {
 #endif
 
+/*
+#define _JIT_COMPILE_DEBUG	1
+*/
+
+
 /*
  * Determine what kind of Win32 system we are running on.
  */
@@ -219,6 +224,7 @@ struct _jit_value
 	jit_nint			address;
 	jit_nint			frame_offset;
 	jit_nuint			usage_count;
+	int				index;
 };
 #define	JIT_INVALID_FRAME_OFFSET	((jit_nint)0x7FFFFFFF)
 
@@ -316,6 +322,9 @@ struct _jit_builder
 	/* Flag that indicates that the current function contains a tail call */
 	int					has_tail_call : 1;
 
+	/* Generate position-independent code */
+	int					position_independent : 1;
+
 	/* List of all instructions in this function */
 	jit_insn_t		   *insns;
 	int					num_insns;
@@ -347,6 +356,10 @@ struct _jit_builder
 	/* Size of the outgoing parameter area in the frame */
 	jit_nint			param_area_size;
 
+#ifdef _JIT_COMPILE_DEBUG
+	int				block_count;
+	int				insn_count;
+#endif
 };
 
 /*
@@ -382,17 +395,20 @@ struct _jit_function
 	/* The entry point for the function's compiled code */
 	void * volatile		entry_point;
 
-	/* The closure/vtable entry point for the function's compiled code */
-	void * volatile		closure_entry;
-
 	/* The function to call to perform on-demand compilation */
 	jit_on_demand_func	on_demand;
 
 #ifdef jit_redirector_size
 	/* Buffer that contains the redirector for this function.
 	   Redirectors are used to support on-demand compilation */
-	unsigned char		redirector[jit_redirector_size];
+	unsigned char		*redirector;
 #endif
+
+	/* Buffer that contains the indirector for this function.
+	   The indirector jumps to the address that is currently
+	   stored in the entry_point field. Indirectors are used
+	   to support recompilation and on-demand compilation. */
+	unsigned char		*indirector;
 };
 
 /*
@@ -462,6 +478,9 @@ struct _jit_context
 	/* Debugger support */
 	jit_debugger_hook_func debug_hook;
 	jit_debugger_t		debugger;
+
+	/* On-demand compilation driver */
+	jit_on_demand_driver_func	on_demand_driver;
 };
 
 /*
@@ -635,6 +654,15 @@ extern struct _jit_type const _jit_type_void_ptr_def;
  */
 #define	JIT_CALL_NATIVE		(1 << 14)
 
+#ifdef JIT_USE_SIGNALS
+
+/*
+ * Initialize the signal handlers.
+ */
+void _jit_signal_init(void);
+
+#endif
+
 #ifdef	__cplusplus
 };
 #endif
