@@ -536,6 +536,10 @@ jit_opcode_info_t const jit_opcodes[JIT_OP_NUM_OPCODES] = {
 	{"mark_offset",					F_(EMPTY, INT, EMPTY)},
 	{"mark_breakpoint",				F_(EMPTY, PTR, PTR)},
 
+	/*
+	 * Switch statement support.
+	 */
+	{"jump_table",				F_(ANY, PTR, INT)|JIT_OPCODE_IS_JUMP_TABLE},
 };
 
 #if defined(JIT_BACKEND_INTERP)
@@ -545,86 +549,130 @@ jit_opcode_info_t const _jit_interp_opcodes[JIT_OP_NUM_INTERP_OPCODES] = {
 	/*
 	 * Argument variable access opcodes.
 	 */
-	{"ldarg_sbyte",					JIT_OPCODE_NINT_ARG},
-	{"ldarg_ubyte",					JIT_OPCODE_NINT_ARG},
-	{"ldarg_short",					JIT_OPCODE_NINT_ARG},
-	{"ldarg_ushort",				JIT_OPCODE_NINT_ARG},
-	{"ldarg_int",					JIT_OPCODE_NINT_ARG},
-	{"ldarg_long",					JIT_OPCODE_NINT_ARG},
-	{"ldarg_float32",				JIT_OPCODE_NINT_ARG},
-	{"ldarg_float64",				JIT_OPCODE_NINT_ARG},
-	{"ldarg_nfloat",				JIT_OPCODE_NINT_ARG},
-	{"ldarg_struct",				JIT_OPCODE_NINT_ARG_TWO},
-	{"ldarga",						JIT_OPCODE_NINT_ARG},
-	{"starg_byte",					JIT_OPCODE_NINT_ARG},
-	{"starg_short",					JIT_OPCODE_NINT_ARG},
-	{"starg_int",					JIT_OPCODE_NINT_ARG},
-	{"starg_long",					JIT_OPCODE_NINT_ARG},
-	{"starg_float32",				JIT_OPCODE_NINT_ARG},
-	{"starg_float64",				JIT_OPCODE_NINT_ARG},
-	{"starg_nfloat",				JIT_OPCODE_NINT_ARG},
-	{"starg_struct",				JIT_OPCODE_NINT_ARG_TWO},
+	{"lda_0_sbyte",			JIT_OPCODE_NINT_ARG},
+	{"lda_0_ubyte",			JIT_OPCODE_NINT_ARG},
+	{"lda_0_short",			JIT_OPCODE_NINT_ARG},
+	{"lda_0_ushort",		JIT_OPCODE_NINT_ARG},
+	{"lda_0_int",			JIT_OPCODE_NINT_ARG},
+	{"lda_0_long",			JIT_OPCODE_NINT_ARG},
+	{"lda_0_float32",		JIT_OPCODE_NINT_ARG},
+	{"lda_0_float64",		JIT_OPCODE_NINT_ARG},
+	{"lda_0_nfloat",		JIT_OPCODE_NINT_ARG},
+	{"ldaa_0",			JIT_OPCODE_NINT_ARG},
+	{"lda_1_sbyte",			JIT_OPCODE_NINT_ARG},
+	{"lda_1_ubyte",			JIT_OPCODE_NINT_ARG},
+	{"lda_1_short",			JIT_OPCODE_NINT_ARG},
+	{"lda_1_ushort",		JIT_OPCODE_NINT_ARG},
+	{"lda_1_int",			JIT_OPCODE_NINT_ARG},
+	{"lda_1_long",			JIT_OPCODE_NINT_ARG},
+	{"lda_1_float32",		JIT_OPCODE_NINT_ARG},
+	{"lda_1_float64",		JIT_OPCODE_NINT_ARG},
+	{"lda_1_nfloat",		JIT_OPCODE_NINT_ARG},
+	{"ldaa_1",			JIT_OPCODE_NINT_ARG},
+	{"lda_2_sbyte",			JIT_OPCODE_NINT_ARG},
+	{"lda_2_ubyte",			JIT_OPCODE_NINT_ARG},
+	{"lda_2_short",			JIT_OPCODE_NINT_ARG},
+	{"lda_2_ushort",		JIT_OPCODE_NINT_ARG},
+	{"lda_2_int",			JIT_OPCODE_NINT_ARG},
+	{"lda_2_long",			JIT_OPCODE_NINT_ARG},
+	{"lda_2_float32",		JIT_OPCODE_NINT_ARG},
+	{"lda_2_float64",		JIT_OPCODE_NINT_ARG},
+	{"lda_2_nfloat",		JIT_OPCODE_NINT_ARG},
+	{"ldaa_2",			JIT_OPCODE_NINT_ARG},
+	{"sta_0_byte",			JIT_OPCODE_NINT_ARG},
+	{"sta_0_short",			JIT_OPCODE_NINT_ARG},
+	{"sta_0_int",			JIT_OPCODE_NINT_ARG},
+	{"sta_0_long",			JIT_OPCODE_NINT_ARG},
+	{"sta_0_float32",		JIT_OPCODE_NINT_ARG},
+	{"sta_0_float64",		JIT_OPCODE_NINT_ARG},
+	{"sta_0_nfloat",		JIT_OPCODE_NINT_ARG},
 
 	/*
 	 * Local variable frame access opcodes.
 	 */
-	{"ldloc_sbyte",					JIT_OPCODE_NINT_ARG},
-	{"ldloc_ubyte",					JIT_OPCODE_NINT_ARG},
-	{"ldloc_short",					JIT_OPCODE_NINT_ARG},
-	{"ldloc_ushort",				JIT_OPCODE_NINT_ARG},
-	{"ldloc_int",					JIT_OPCODE_NINT_ARG},
-	{"ldloc_long",					JIT_OPCODE_NINT_ARG},
-	{"ldloc_float32",				JIT_OPCODE_NINT_ARG},
-	{"ldloc_float64",				JIT_OPCODE_NINT_ARG},
-	{"ldloc_nfloat",				JIT_OPCODE_NINT_ARG},
-	{"ldloc_struct",				JIT_OPCODE_NINT_ARG_TWO},
-	{"ldloca",						JIT_OPCODE_NINT_ARG},
-	{"stloc_byte",					JIT_OPCODE_NINT_ARG},
-	{"stloc_short",					JIT_OPCODE_NINT_ARG},
-	{"stloc_int",					JIT_OPCODE_NINT_ARG},
-	{"stloc_long",					JIT_OPCODE_NINT_ARG},
-	{"stloc_float32",				JIT_OPCODE_NINT_ARG},
-	{"stloc_float64",				JIT_OPCODE_NINT_ARG},
-	{"stloc_nfloat",				JIT_OPCODE_NINT_ARG},
-	{"stloc_struct",				JIT_OPCODE_NINT_ARG_TWO},
+	{"ldl_0_sbyte",			JIT_OPCODE_NINT_ARG},
+	{"ldl_0_ubyte",			JIT_OPCODE_NINT_ARG},
+	{"ldl_0_short",			JIT_OPCODE_NINT_ARG},
+	{"ldl_0_ushort",		JIT_OPCODE_NINT_ARG},
+	{"ldl_0_int",			JIT_OPCODE_NINT_ARG},
+	{"ldl_0_long",			JIT_OPCODE_NINT_ARG},
+	{"ldl_0_float32",		JIT_OPCODE_NINT_ARG},
+	{"ldl_0_float64",		JIT_OPCODE_NINT_ARG},
+	{"ldl_0_nfloat",		JIT_OPCODE_NINT_ARG},
+	{"ldla_0",			JIT_OPCODE_NINT_ARG},
+	{"ldl_1_sbyte",			JIT_OPCODE_NINT_ARG},
+	{"ldl_1_ubyte",			JIT_OPCODE_NINT_ARG},
+	{"ldl_1_short",			JIT_OPCODE_NINT_ARG},
+	{"ldl_1_ushort",		JIT_OPCODE_NINT_ARG},
+	{"ldl_1_int",			JIT_OPCODE_NINT_ARG},
+	{"ldl_1_long",			JIT_OPCODE_NINT_ARG},
+	{"ldl_1_float32",		JIT_OPCODE_NINT_ARG},
+	{"ldl_1_float64",		JIT_OPCODE_NINT_ARG},
+	{"ldl_1_nfloat",		JIT_OPCODE_NINT_ARG},
+	{"ldla_1",			JIT_OPCODE_NINT_ARG},
+	{"ldl_2_sbyte",			JIT_OPCODE_NINT_ARG},
+	{"ldl_2_ubyte",			JIT_OPCODE_NINT_ARG},
+	{"ldl_2_short",			JIT_OPCODE_NINT_ARG},
+	{"ldl_2_ushort",		JIT_OPCODE_NINT_ARG},
+	{"ldl_2_int",			JIT_OPCODE_NINT_ARG},
+	{"ldl_2_long",			JIT_OPCODE_NINT_ARG},
+	{"ldl_2_float32",		JIT_OPCODE_NINT_ARG},
+	{"ldl_2_float64",		JIT_OPCODE_NINT_ARG},
+	{"ldl_2_nfloat",		JIT_OPCODE_NINT_ARG},
+	{"ldla_2",			JIT_OPCODE_NINT_ARG},
+	{"stl_0_byte",			JIT_OPCODE_NINT_ARG},
+	{"stl_0_short",			JIT_OPCODE_NINT_ARG},
+	{"stl_0_int",			JIT_OPCODE_NINT_ARG},
+	{"stl_0_long",			JIT_OPCODE_NINT_ARG},
+	{"stl_0_float32",		JIT_OPCODE_NINT_ARG},
+	{"stl_0_float64",		JIT_OPCODE_NINT_ARG},
+	{"stl_0_nfloat",		JIT_OPCODE_NINT_ARG},
 
 	/*
-	 * Pointer check opcodes (interpreter only).
+	 * Load constant values.
 	 */
-	{"check_null_n",				JIT_OPCODE_NINT_ARG},
+	{"ldc_0_int",			JIT_OPCODE_NINT_ARG},
+	{"ldc_1_int",			JIT_OPCODE_NINT_ARG},
+	{"ldc_2_int",			JIT_OPCODE_NINT_ARG},
+	{"ldc_0_long",			JIT_OPCODE_CONST_LONG},
+	{"ldc_1_long",			JIT_OPCODE_CONST_LONG},
+	{"ldc_2_long",			JIT_OPCODE_CONST_LONG},
+	{"ldc_0_float32",		JIT_OPCODE_CONST_FLOAT32},
+	{"ldc_1_float32",		JIT_OPCODE_CONST_FLOAT32},
+	{"ldc_2_float32",		JIT_OPCODE_CONST_FLOAT32},
+	{"ldc_0_float64",		JIT_OPCODE_CONST_FLOAT64},
+	{"ldc_1_float64",		JIT_OPCODE_CONST_FLOAT64},
+	{"ldc_2_float64",		JIT_OPCODE_CONST_FLOAT64},
+	{"ldc_0_nfloat",		JIT_OPCODE_CONST_NFLOAT},
+	{"ldc_1_nfloat",		JIT_OPCODE_CONST_NFLOAT},
+	{"ldc_2_nfloat",		JIT_OPCODE_CONST_NFLOAT},
+
+	/*
+	 * Load return value.
+	 */
+	{"ldr_0_int",			0},
+	{"ldr_0_long",			0},
+	{"ldr_0_float32",		0},
+	{"ldr_0_float64",		0},
+	{"ldr_0_nfloat",		0},
 
 	/*
 	 * Stack management.
 	 */
-	{"pop",							0},
-	{"pop_2",						0},
-	{"pop_3",						0},
-	{"push_return_int",				0},
-	{"push_return_long",			0},
-	{"push_return_float32",			0},
-	{"push_return_float64",			0},
-	{"push_return_nfloat",			0},
-	{"push_return_small_struct",	JIT_OPCODE_NINT_ARG},
+	{"pop",				0},
+	{"pop_2",			0},
+	{"pop_3",			0},
 
 	/*
 	 * Nested function call handling.
 	 */
-	{"import_local",				JIT_OPCODE_NINT_ARG_TWO},
-	{"import_arg",					JIT_OPCODE_NINT_ARG_TWO},
-
-	/*
-	 * Push constant values onto the stack.
-	 */
-	{"push_const_int",				JIT_OPCODE_NINT_ARG},
-	{"push_const_long",				JIT_OPCODE_CONST_LONG},
-	{"push_const_float32",			JIT_OPCODE_CONST_FLOAT32},
-	{"push_const_float64",			JIT_OPCODE_CONST_FLOAT64},
-	{"push_const_nfloat",			JIT_OPCODE_CONST_NFLOAT},
+	{"import_local",		JIT_OPCODE_NINT_ARG_TWO},
+	{"import_arg",			JIT_OPCODE_NINT_ARG_TWO},
 
 	/*
 	 * Marker opcode for the end of a function.
 	 */
-	{"end_marker",					0},
+	{"end_marker",			0},
 };
 
 #endif /* JIT_BACKEND_INTERP */
