@@ -761,7 +761,12 @@ void _jit_gen_load_value
 	jit_cache_end_output();
 }
 
-void _jit_gen_load_global(jit_gencode_t gen, jit_value_t value)
+void _jit_gen_spill_global(jit_gencode_t gen, int reg, jit_value_t value)
+{
+	/* TODO: Implement if ARM needs it. */
+}
+
+void _jit_gen_load_global(jit_gencode_t gen, int reg, jit_value_t value)
 {
 	jit_cache_setup_output(32);
 	arm_load_membase(inst, _jit_reg_info[value->global_reg].cpu_reg,
@@ -769,6 +774,18 @@ void _jit_gen_load_global(jit_gencode_t gen, jit_value_t value)
 	jit_cache_end_output();
 }
 
+void _jit_gen_exch_top(jit_gencode_t gen, int reg)
+{
+}
+
+void _jit_gen_move_top(jit_gencode_t gen, int reg)
+{
+}
+
+void _jit_gen_spill_top(jit_gencode_t gen, int reg, jit_value_t value, int pop)
+{
+}
+
 void _jit_gen_fix_value(jit_value_t value)
 {
 	if(!(value->has_frame_offset) && !(value->is_constant))
