@@ -44,7 +44,7 @@
 #define	X86_REG_ESP			7
 #define	X86_REG_ST0			8
 #define	X86_REG_ST1			9
-#define	X85_REG_ST2			10
+#define	X86_REG_ST2			10
 #define	X86_REG_ST3			11
 #define	X86_REG_ST4			12
 #define	X86_REG_ST5			13
@@ -63,9 +63,34 @@
 #define	ROUND_STACK(size)	\
 		(((size) + (sizeof(void *) - 1)) & ~(sizeof(void *) - 1))
 
+static _jit_regclass_t *x86_reg;
+static _jit_regclass_t *x86_breg;
+static _jit_regclass_t *x86_freg;
+static _jit_regclass_t *x86_lreg;
+
 void _jit_init_backend(void)
 {
-	/* Nothing to do here for the x86 */
+	x86_reg = _jit_regclass_create(
+		"reg", JIT_REG_WORD, 6,
+		X86_REG_EAX, X86_REG_ECX,
+		X86_REG_EDX, X86_REG_EBX,
+		X86_REG_ESI, X86_REG_EDI);
+
+	x86_breg = _jit_regclass_create(
+		"breg", JIT_REG_WORD, 4,
+		X86_REG_EAX, X86_REG_ECX,
+		X86_REG_EDX, X86_REG_EBX);
+
+	x86_freg = _jit_regclass_create(
+		"freg", JIT_REG_X86_FLOAT | JIT_REG_IN_STACK, 8,
+		X86_REG_ST0, X86_REG_ST1,
+		X86_REG_ST2, X86_REG_ST3,
+		X86_REG_ST4, X86_REG_ST5,
+		X86_REG_ST6, X86_REG_ST7);
+
+	x86_lreg = _jit_regclass_create(
+		"lreg", JIT_REG_LONG, 2,
+		X86_REG_EAX, X86_REG_ECX);
 }
 
 void _jit_gen_get_elf_info(jit_elf_info_t *info)
@@ -149,7 +174,7 @@ int _jit_create_call_return_insns
 	/* Bail out now if we don't need to worry about return values */
 	if(!return_value || ptr_return)
 	{
-		return 0;
+		return 1;
 	}
 
 	/* Structure values must be flushed into the frame, and
@@ -186,9 +211,9 @@ int _jit_opcode_is_supported(int opcode)
 {
 	switch(opcode)
 	{
-		#define JIT_INCLUDE_SUPPORTED
-		#include "jit-rules-x86.slc"
-		#undef JIT_INCLUDE_SUPPORTED
+	#define JIT_INCLUDE_SUPPORTED
+	#include "jit-rules-x86.inc"
+	#undef JIT_INCLUDE_SUPPORTED
 	}
 	return 0;
 }
@@ -373,6 +398,11 @@ void _jit_gen_epilog(jit_gencode_t gen, jit_function_t func)
 	gen->posn.ptr = inst;
 }
 
+#if 0
+/*
+ * The x86 backend does not need this function because it uses
+ * _jit_create_indirector() instead.
+ */
 void *_jit_gen_redirector(jit_gencode_t gen, jit_function_t func)
 {
 	void *ptr, *entry;
@@ -386,6 +416,7 @@ void *_jit_gen_redirector(jit_gencode_t gen, jit_function_t func)
 	x86_jump_mem(gen->posn.ptr, ptr);
 	return entry;
 }
+#endif
 
 /*
  * Setup or teardown the x86 code output process.
@@ -428,140 +459,6 @@ static int get_temp_reg(int reg1, int reg2, int reg3)
 	return X86_EDI;
 }
 
-/*
- * Load a small structure from a pointer into registers.
- */
-static unsigned char *load_small_struct
-	(unsigned char *inst, int reg, int other_reg,
-	 int base_reg, jit_nint offset, jit_nint size, int save_temp)
-{
-	int temp_reg;
-	switch(size)
-	{
-		case 1:
-		{
-			x86_widen_membase(inst, reg, base_reg, offset, 0, 0);
-		}
-		break;
-
-		case 2:
-		{
-			x86_widen_membase(inst, reg, base_reg, offset, 0, 1);
-		}
-		break;
-
-		case 3:
-		{
-			temp_reg = get_temp_reg(reg, -1, base_reg);
-			if(save_temp || temp_reg >= X86_EBX)
-			{
-				x86_push_reg(inst, temp_reg);
-			}
-			x86_widen_membase(inst, temp_reg, base_reg, offset, 0, 1);
-			x86_widen_membase(inst, reg, base_reg, offset + 2, 0, 0);
-			x86_shift_reg_imm(inst, X86_SHL, reg, 16);
-			x86_alu_reg_reg(inst, X86_OR, reg, temp_reg);
-			if(save_temp || temp_reg >= X86_EBX)
-			{
-				x86_pop_reg(inst, temp_reg);
-			}
-		}
-		break;
-
-		case 4:
-		{
-			x86_mov_reg_membase(inst, reg, base_reg, offset, 4);
-		}
-		break;
-
-		case 5:
-		{
-			if(reg != base_reg)
-			{
-				x86_mov_reg_membase(inst, reg, base_reg, offset, 4);
-				x86_widen_membase(inst, other_reg, base_reg, offset + 4, 0, 0);
-			}
-			else
-			{
-				x86_widen_membase(inst, other_reg, base_reg, offset + 4, 0, 0);
-				x86_mov_reg_membase(inst, reg, base_reg, offset, 4);
-			}
-		}
-		break;
-
-		case 6:
-		{
-			if(reg != base_reg)
-			{
-				x86_mov_reg_membase(inst, reg, base_reg, offset, 4);
-				x86_widen_membase(inst, other_reg, base_reg, offset + 4, 0, 1);
-			}
-			else
-			{
-				x86_widen_membase(inst, other_reg, base_reg, offset + 4, 0, 1);
-				x86_mov_reg_membase(inst, reg, base_reg, offset, 4);
-			}
-		}
-		break;
-
-		case 7:
-		{
-			temp_reg = get_temp_reg(reg, other_reg, base_reg);
-			if(save_temp || temp_reg >= X86_EBX)
-			{
-				x86_push_reg(inst, temp_reg);
-			}
-			if(reg != base_reg && other_reg != base_reg)
-			{
-				x86_mov_reg_membase(inst, reg, base_reg, offset, 4);
-				x86_widen_membase(inst, other_reg, base_reg, offset + 4, 0, 1);
-				x86_widen_membase(inst, temp_reg, base_reg, offset + 6, 0, 0);
-				x86_shift_reg_imm(inst, X86_SHL, temp_reg, 16);
-				x86_alu_reg_reg(inst, X86_OR, other_reg, temp_reg);
-			}
-			else if(reg != base_reg)
-			{
-				/* other_reg == base_reg */
-				x86_mov_reg_membase(inst, reg, base_reg, offset, 4);
-				x86_widen_membase(inst, temp_reg, base_reg, offset + 6, 0, 0);
-				x86_widen_membase(inst, other_reg, base_reg, offset + 4, 0, 1);
-				x86_shift_reg_imm(inst, X86_SHL, temp_reg, 16);
-				x86_alu_reg_reg(inst, X86_OR, other_reg, temp_reg);
-			}
-			else
-			{
-				/* reg == base_reg */
-				x86_widen_membase(inst, other_reg, base_reg, offset + 4, 0, 1);
-				x86_widen_membase(inst, temp_reg, base_reg, offset + 6, 0, 0);
-				x86_shift_reg_imm(inst, X86_SHL, temp_reg, 16);
-				x86_alu_reg_reg(inst, X86_OR, other_reg, temp_reg);
-				x86_mov_reg_membase(inst, reg, base_reg, offset, 4);
-			}
-			if(save_temp || temp_reg >= X86_EBX)
-			{
-				x86_pop_reg(inst, temp_reg);
-			}
-		}
-		break;
-
-		case 8:
-		{
-			if(reg != base_reg)
-			{
-				x86_mov_reg_membase(inst, reg, base_reg, offset, 4);
-				x86_mov_reg_membase(inst, other_reg, base_reg, offset + 4, 4);
-			}
-			else
-			{
-				x86_mov_reg_membase(inst, other_reg, base_reg, offset + 4, 4);
-				x86_mov_reg_membase(inst, reg, base_reg, offset, 4);
-			}
-		}
-		break;
-	}
-	return inst;
-}
-
 /*
  * Store a byte value to a membase address.
  */
@@ -760,19 +657,97 @@ void _jit_gen_free_reg(jit_gencode_t gen, int reg,
 	}
 }
 
-void _jit_gen_load_value
-	(jit_gencode_t gen, int reg, int other_reg, jit_value_t value)
+static int
+fp_stack_index(jit_gencode_t gen, int reg)
 {
+	return gen->reg_stack_top - reg - 1;
+}
+
+void
+_jit_gen_exch_top(jit_gencode_t gen, int reg)
+{
+	if(IS_FLOAT_REG(reg))
+	{
+		jit_cache_setup_output(2);
+		x86_fxch(inst, fp_stack_index(gen, reg));
+		jit_cache_end_output();
+	}
+}
+
+void
+ _jit_gen_move_top(jit_gencode_t gen, int reg)
+{
+	if(IS_FLOAT_REG(reg))
+	{
+		jit_cache_setup_output(2);
+		x86_fstp(inst, fp_stack_index(gen, reg));
+		jit_cache_end_output();
+	}
+}
+
+void
+_jit_gen_spill_top(jit_gencode_t gen, int reg, jit_value_t value, int pop)
+{
+	int offset;
+	if(IS_FLOAT_REG(reg))
+	{
+		/* Make sure that we have sufficient space */
+		jit_cache_setup_output(16);
+
+		/* Fix the value in place within the local variable frame */
+		_jit_gen_fix_value(value);
+
+		/* Output an appropriate instruction to spill the value */
+		offset = (int)(value->frame_offset);
+
+		/* Spill the top of the floating-point register stack */
+		switch(jit_type_normalize(value->type)->kind)
+		{
+			case JIT_TYPE_FLOAT32:
+			{
+				x86_fst_membase(inst, X86_EBP, offset, 0, pop);
+			}
+			break;
+
+			case JIT_TYPE_FLOAT64:
+			{
+				x86_fst_membase(inst, X86_EBP, offset, 1, pop);
+			}
+			break;
+
+			case JIT_TYPE_NFLOAT:
+			{
+				x86_fst80_membase(inst, X86_EBP, offset);
+				if(!pop)
+				{
+					x86_fld80_membase(inst, X86_EBP, offset);
+				}
+			}
+			break;
+		}
+
+		/* End the code output process */
+		jit_cache_end_output();
+	}
+}
+
+void
+_jit_gen_load_value(jit_gencode_t gen, int reg, int other_reg, jit_value_t value)
+{
+	jit_type_t type;
+	int src_reg, other_src_reg;
 	void *ptr;
 	int offset;
 
 	/* Make sure that we have sufficient space */
 	jit_cache_setup_output(16);
 
+	type = jit_type_normalize(value->type);
+
+	/* Load zero */
 	if(value->is_constant)
 	{
-		/* Determine the type of constant to be loaded */
-		switch(jit_type_normalize(value->type)->kind)
+		switch(type->kind)
 		{
 			case JIT_TYPE_SBYTE:
 			case JIT_TYPE_UBYTE:
@@ -781,8 +756,15 @@ void _jit_gen_load_value
 			case JIT_TYPE_INT:
 			case JIT_TYPE_UINT:
 			{
-				x86_mov_reg_imm(inst, _jit_reg_info[reg].cpu_reg,
-								(jit_nint)(value->address));
+				if((jit_nint)(value->address) == 0)
+				{
+					x86_clear_reg(inst, _jit_reg_info[reg].cpu_reg);
+				}
+				else
+				{
+					x86_mov_reg_imm(inst, _jit_reg_info[reg].cpu_reg,
+							(jit_nint)(value->address));
+				}
 			}
 			break;
 
@@ -791,15 +773,27 @@ void _jit_gen_load_value
 			{
 				jit_long long_value;
 				long_value = jit_value_get_long_constant(value);
-			#ifdef JIT_NATIVE_INT64
-				x86_mov_reg_imm(inst, _jit_reg_info[reg].cpu_reg,
-								(jit_nint)long_value);
-			#else
-				x86_mov_reg_imm(inst, _jit_reg_info[reg].cpu_reg,
-								(jit_int)long_value);
-				x86_mov_reg_imm(inst, _jit_reg_info[other_reg].cpu_reg,
-								(jit_int)(long_value >> 32));
-			#endif
+				if(long_value == 0)
+				{
+#ifdef JIT_NATIVE_INT64
+					x86_clear_reg(inst, _jit_reg_info[reg].cpu_reg);
+#else
+					x86_clear_reg(inst, _jit_reg_info[reg].cpu_reg);
+					x86_clear_reg(inst, _jit_reg_info[other_reg].cpu_reg);
+#endif
+				}
+				else
+				{
+#ifdef JIT_NATIVE_INT64
+					x86_mov_reg_imm(inst, _jit_reg_info[reg].cpu_reg,
+							(jit_nint)long_value);
+#else
+					x86_mov_reg_imm(inst, _jit_reg_info[reg].cpu_reg,
+							(jit_int)long_value);
+					x86_mov_reg_imm(inst, _jit_reg_info[other_reg].cpu_reg,
+							(jit_int)(long_value >> 32));
+#endif
+				}
 			}
 			break;
 
@@ -807,21 +801,27 @@ void _jit_gen_load_value
 			{
 				jit_float32 float32_value;
 				float32_value = jit_value_get_float32_constant(value);
-				if(!jit_cache_check_for_n(&(gen->posn), 32))
-				{
-					jit_cache_mark_full(&(gen->posn));
-					return;
-				}
 				if(IS_WORD_REG(reg))
 				{
 					x86_mov_reg_imm(inst, _jit_reg_info[reg].cpu_reg,
-									((jit_int *)&float32_value)[0]);
+							((jit_int *)&float32_value)[0]);
 				}
 				else
 				{
-					ptr = _jit_cache_alloc(&(gen->posn), sizeof(jit_float32));
-					jit_memcpy(ptr, &float32_value, sizeof(float32_value));
-					x86_fld(inst, ptr, 0);
+					if(float32_value == (jit_float32) 0.0)
+					{
+						x86_fldz(inst);
+					}
+					else if(float32_value == (jit_float32) 1.0)
+					{
+						x86_fld1(inst);
+					}
+					else
+					{
+						ptr = _jit_cache_alloc(&(gen->posn), sizeof(jit_float32));
+						jit_memcpy(ptr, &float32_value, sizeof(float32_value));
+						x86_fld(inst, ptr, 0);
+					}
 				}
 			}
 			break;
@@ -830,23 +830,29 @@ void _jit_gen_load_value
 			{
 				jit_float64 float64_value;
 				float64_value = jit_value_get_float64_constant(value);
-				if(!jit_cache_check_for_n(&(gen->posn), 32))
-				{
-					jit_cache_mark_full(&(gen->posn));
-					return;
-				}
 				if(IS_WORD_REG(reg))
 				{
 					x86_mov_reg_imm(inst, _jit_reg_info[reg].cpu_reg,
-									((jit_int *)&float64_value)[0]);
+							((jit_int *)&float64_value)[0]);
 					x86_mov_reg_imm(inst, _jit_reg_info[other_reg].cpu_reg,
-									((jit_int *)&float64_value)[1]);
+							((jit_int *)&float64_value)[1]);
 				}
 				else
 				{
-					ptr = _jit_cache_alloc(&(gen->posn), sizeof(jit_float64));
-					jit_memcpy(ptr, &float64_value, sizeof(float64_value));
-					x86_fld(inst, ptr, 1);
+					if(float64_value == (jit_float64) 0.0)
+					{
+						x86_fldz(inst);
+					}
+					else if(float64_value == (jit_float64) 1.0)
+					{
+						x86_fld1(inst);
+					}
+					else
+					{
+						ptr = _jit_cache_alloc(&(gen->posn), sizeof(jit_float64));
+						jit_memcpy(ptr, &float64_value, sizeof(float64_value));
+						x86_fld(inst, ptr, 1);
+					}
 				}
 			}
 			break;
@@ -855,42 +861,131 @@ void _jit_gen_load_value
 			{
 				jit_nfloat nfloat_value;
 				nfloat_value = jit_value_get_nfloat_constant(value);
-				if(!jit_cache_check_for_n(&(gen->posn), 32))
-				{
-					jit_cache_mark_full(&(gen->posn));
-					return;
-				}
-				if(IS_WORD_REG(reg) &&
-				   sizeof(jit_nfloat) == sizeof(jit_float64))
+				if(IS_WORD_REG(reg) && sizeof(jit_nfloat) == sizeof(jit_float64))
 				{
 					x86_mov_reg_imm(inst, _jit_reg_info[reg].cpu_reg,
-									((jit_int *)&nfloat_value)[0]);
+							((jit_int *)&nfloat_value)[0]);
 					x86_mov_reg_imm(inst, _jit_reg_info[other_reg].cpu_reg,
-									((jit_int *)&nfloat_value)[1]);
+							((jit_int *)&nfloat_value)[1]);
 				}
 				else
 				{
-					ptr = _jit_cache_alloc(&(gen->posn), sizeof(jit_nfloat));
-					jit_memcpy(ptr, &nfloat_value, sizeof(nfloat_value));
-					if(sizeof(jit_nfloat) != sizeof(jit_float64))
+					if(nfloat_value == (jit_nfloat) 0.0)
 					{
-						x86_fld80_mem(inst, ptr);
+						x86_fldz(inst);
+					}
+					else if(nfloat_value == (jit_nfloat) 1.0)
+					{
+						x86_fld1(inst);
 					}
 					else
 					{
-						x86_fld(inst, ptr, 1);
+						ptr = _jit_cache_alloc(&(gen->posn), sizeof(jit_nfloat));
+						jit_memcpy(ptr, &nfloat_value, sizeof(nfloat_value));
+						if(sizeof(jit_nfloat) == sizeof(jit_float64))
+						{
+							x86_fld(inst, ptr, 1);
+						}
+						else
+						{
+							x86_fld80_mem(inst, ptr);
+						}
 					}
 				}
 			}
 			break;
 		}
 	}
-	else if(value->has_global_register)
+	else if(value->in_register || value->in_global_register)
 	{
-		/* Load the value out of a global register */
-		x86_mov_reg_reg(inst, _jit_reg_info[reg].cpu_reg,
-						_jit_reg_info[value->global_reg].cpu_reg,
-						sizeof(void *));
+		if(value->in_register)
+		{
+			src_reg = value->reg;
+			if(other_reg >= 0)
+			{
+				other_src_reg = _jit_reg_info[src_reg].other_reg;
+			}
+			else
+			{
+				other_src_reg = -1;
+			}
+		}
+		else
+		{
+			src_reg = value->global_reg;
+			other_src_reg = -1;
+		}
+
+		switch(type->kind)
+		{
+#if 0
+			case JIT_TYPE_SBYTE:
+			{
+				x86_widen_reg(inst, _jit_reg_info[reg].cpu_reg,
+					      _jit_reg_info[src_reg].cpu_reg, 1, 0);
+			}
+			break;
+
+			case JIT_TYPE_UBYTE:
+			{
+				x86_widen_reg(inst, _jit_reg_info[reg].cpu_reg,
+					      _jit_reg_info[src_reg].cpu_reg, 0, 0);
+			}
+			break;
+
+			case JIT_TYPE_SHORT:
+			{
+				x86_widen_reg(inst, _jit_reg_info[reg].cpu_reg,
+					      _jit_reg_info[src_reg].cpu_reg, 1, 1);
+			}
+			break;
+
+			case JIT_TYPE_USHORT:
+			{
+				x86_widen_reg(inst, _jit_reg_info[reg].cpu_reg,
+					      _jit_reg_info[src_reg].cpu_reg, 0, 1);
+			}
+			break;
+#else
+			case JIT_TYPE_SBYTE:
+			case JIT_TYPE_UBYTE:
+			case JIT_TYPE_SHORT:
+			case JIT_TYPE_USHORT:
+#endif
+			case JIT_TYPE_INT:
+			case JIT_TYPE_UINT:
+			{
+				x86_mov_reg_reg(inst, _jit_reg_info[reg].cpu_reg,
+						_jit_reg_info[src_reg].cpu_reg, 4);
+			}
+			break;
+
+			case JIT_TYPE_LONG:
+			case JIT_TYPE_ULONG:
+			{
+#ifdef JIT_NATIVE_INT64
+				x86_mov_reg_reg(inst, _jit_reg_info[reg].cpu_reg,
+						_jit_reg_info[src_reg].cpu_reg, 8);
+#else
+				x86_mov_reg_reg(inst, _jit_reg_info[reg].cpu_reg,
+						_jit_reg_info[src_reg].cpu_reg, 4);
+				x86_mov_reg_reg(inst, _jit_reg_info[other_reg].cpu_reg,
+						_jit_reg_info[other_src_reg].cpu_reg, 4);
+#endif
+			}
+			break;
+
+			case JIT_TYPE_FLOAT32:
+			case JIT_TYPE_FLOAT64:
+			case JIT_TYPE_NFLOAT:
+			{
+				if(!IS_WORD_REG(reg))
+				{
+					x86_fld_reg(inst, fp_stack_index(gen, src_reg));
+				}
+			}
+			break;
+		}
 	}
 	else
 	{
@@ -899,33 +994,33 @@ void _jit_gen_load_value
 		offset = (int)(value->frame_offset);
 
 		/* Load the value into the specified register */
-		switch(jit_type_normalize(value->type)->kind)
+		switch(type->kind)
 		{
 			case JIT_TYPE_SBYTE:
 			{
 				x86_widen_membase(inst, _jit_reg_info[reg].cpu_reg,
-								  X86_EBP, offset, 1, 0);
+						  X86_EBP, offset, 1, 0);
 			}
 			break;
 
 			case JIT_TYPE_UBYTE:
 			{
 				x86_widen_membase(inst, _jit_reg_info[reg].cpu_reg,
-								  X86_EBP, offset, 0, 0);
+						  X86_EBP, offset, 0, 0);
 			}
 			break;
 
 			case JIT_TYPE_SHORT:
 			{
 				x86_widen_membase(inst, _jit_reg_info[reg].cpu_reg,
-								  X86_EBP, offset, 1, 1);
+						  X86_EBP, offset, 1, 1);
 			}
 			break;
 
 			case JIT_TYPE_USHORT:
 			{
 				x86_widen_membase(inst, _jit_reg_info[reg].cpu_reg,
-								  X86_EBP, offset, 0, 1);
+						  X86_EBP, offset, 0, 1);
 			}
 			break;
 
@@ -933,22 +1028,22 @@ void _jit_gen_load_value
 			case JIT_TYPE_UINT:
 			{
 				x86_mov_reg_membase(inst, _jit_reg_info[reg].cpu_reg,
-									X86_EBP, offset, 4);
+						    X86_EBP, offset, 4);
 			}
 			break;
 
 			case JIT_TYPE_LONG:
 			case JIT_TYPE_ULONG:
 			{
-			#ifdef JIT_NATIVE_INT64
+#ifdef JIT_NATIVE_INT64
 				x86_mov_reg_membase(inst, _jit_reg_info[reg].cpu_reg,
-									X86_EBP, offset, 8);
-			#else
+						    X86_EBP, offset, 8);
+#else
 				x86_mov_reg_membase(inst, _jit_reg_info[reg].cpu_reg,
-									X86_EBP, offset, 4);
+						    X86_EBP, offset, 4);
 				x86_mov_reg_membase(inst, _jit_reg_info[other_reg].cpu_reg,
-									X86_EBP, offset + 4, 4);
-			#endif
+						    X86_EBP, offset + 4, 4);
+#endif
 			}
 			break;
 
@@ -957,7 +1052,7 @@ void _jit_gen_load_value
 				if(IS_WORD_REG(reg))
 				{
 					x86_mov_reg_membase(inst, _jit_reg_info[reg].cpu_reg,
-										X86_EBP, offset, 4);
+							    X86_EBP, offset, 4);
 				}
 				else
 				{
@@ -971,9 +1066,9 @@ void _jit_gen_load_value
 				if(IS_WORD_REG(reg))
 				{
 					x86_mov_reg_membase(inst, _jit_reg_info[reg].cpu_reg,
-										X86_EBP, offset, 4);
+							    X86_EBP, offset, 4);
 					x86_mov_reg_membase(inst, _jit_reg_info[other_reg].cpu_reg,
-										X86_EBP, offset + 4, 4);
+							    X86_EBP, offset + 4, 4);
 				}
 				else
 				{
@@ -984,13 +1079,12 @@ void _jit_gen_load_value
 
 			case JIT_TYPE_NFLOAT:
 			{
-				if(IS_WORD_REG(reg) &&
-				   sizeof(jit_nfloat) == sizeof(jit_float64))
+				if(IS_WORD_REG(reg) && sizeof(jit_nfloat) == sizeof(jit_float64))
 				{
 					x86_mov_reg_membase(inst, _jit_reg_info[reg].cpu_reg,
-										X86_EBP, offset, 4);
+							    X86_EBP, offset, 4);
 					x86_mov_reg_membase(inst, _jit_reg_info[other_reg].cpu_reg,
-										X86_EBP, offset + 4, 4);
+							    X86_EBP, offset + 4, 4);
 				}
 				else if(sizeof(jit_nfloat) == sizeof(jit_float64))
 				{
@@ -1009,11 +1103,36 @@ void _jit_gen_load_value
 	jit_cache_end_output();
 }
 
-void _jit_gen_load_global(jit_gencode_t gen, jit_value_t value)
+void _jit_gen_spill_global(jit_gencode_t gen, int reg, jit_value_t value)
 {
 	jit_cache_setup_output(16);
-	x86_mov_reg_membase(inst, _jit_reg_info[value->global_reg].cpu_reg,
-						X86_EBP, value->frame_offset, sizeof(void *));
+	if(value)
+	{
+		_jit_gen_fix_value(value);
+		x86_mov_membase_reg(inst,
+			 X86_EBP, value->frame_offset,
+			_jit_reg_info[value->global_reg].cpu_reg, sizeof(void *));
+	}
+	else
+	{
+		x86_push_reg(inst, _jit_reg_info[reg].cpu_reg);
+	}
+	jit_cache_end_output();
+}
+
+void _jit_gen_load_global(jit_gencode_t gen, int reg, jit_value_t value)
+{
+	jit_cache_setup_output(16);
+	if(value)
+	{
+		x86_mov_reg_membase(inst,
+			_jit_reg_info[value->global_reg].cpu_reg,
+			X86_EBP, value->frame_offset, sizeof(void *));
+	}
+	else
+	{
+		x86_pop_reg(inst, _jit_reg_info[reg].cpu_reg);
+	}
 	jit_cache_end_output();
 }
 
@@ -1028,25 +1147,6 @@ void _jit_gen_fix_value(jit_value_t value)
 	}
 }
 
-/*
- * Widen a byte register.
- */
-static unsigned char *widen_byte(unsigned char *inst, int reg, int isSigned)
-{
-	if(reg == X86_EAX || reg == X86_EBX || reg == X86_ECX || reg == X86_EDX)
-	{
-		x86_widen_reg(inst, reg, reg, isSigned, 0);
-	}
-	else
-	{
-		x86_push_reg(inst, X86_EAX);
-		x86_mov_reg_reg(inst, X86_EAX, reg, 4);
-		x86_widen_reg(inst, reg, X86_EAX, isSigned, 0);
-		x86_pop_reg(inst, X86_EAX);
-	}
-	return inst;
-}
-
 /*
  * Shift the contents of a register.
  */
@@ -1211,43 +1311,6 @@ static unsigned char *jump_to_epilog
 	return inst;
 }
 
-/*
- * Store a byte value to a memindex address.
- */
-static unsigned char *mov_memindex_reg_byte
-		(unsigned char *inst, int basereg,
-		 unsigned offset, int indexreg, int srcreg)
-{
-	if(srcreg == X86_EAX || srcreg == X86_EBX ||
-	   srcreg == X86_ECX || srcreg == X86_EDX)
-	{
-		x86_mov_memindex_reg(inst, basereg, offset, indexreg,
-							 0, srcreg, 1);
-	}
-	else
-	{
-		int tempreg;
-		if(basereg != X86_EAX && indexreg != X86_EAX)
-		{
-			tempreg = X86_EAX;
-		}
-		else if(basereg != X86_ECX && indexreg != X86_ECX)
-		{
-			tempreg = X86_ECX;
-		}
-		else
-		{
-			tempreg = X86_EDX;
-		}
-		x86_push_reg(inst, tempreg);
-		x86_mov_reg_reg(inst, tempreg, srcreg, 4);
-		x86_mov_memindex_reg(inst, basereg, offset, indexreg,
-							 0, tempreg, 1);
-		x86_pop_reg(inst, tempreg);
-	}
-	return inst;
-}
-
 /*
  * Throw a builtin exception.
  */
@@ -1258,10 +1321,20 @@ static unsigned char *throw_builtin
 	if(func->builder->setjmp_value != 0)
 	{
 		_jit_gen_fix_value(func->builder->setjmp_value);
-		x86_call_imm(inst, 0);
-		x86_pop_membase(inst, X86_EBP,
-						func->builder->setjmp_value->frame_offset +
-						jit_jmp_catch_pc_offset);
+		if(func->builder->position_independent)
+		{
+			x86_call_imm(inst, 0);
+			x86_pop_membase(inst, X86_EBP,
+					func->builder->setjmp_value->frame_offset
+					+ jit_jmp_catch_pc_offset);
+		}
+		else
+		{
+			int pc = (int) inst;
+			x86_mov_membase_imm(inst, X86_EBP,
+					    func->builder->setjmp_value->frame_offset
+					    + jit_jmp_catch_pc_offset, pc, 4);
+		}
 	}
 
 	/* Push the exception type onto the stack */
@@ -1366,14 +1439,14 @@ void _jit_gen_insn(jit_gencode_t gen, jit_function_t func,
 {
 	switch(insn->opcode)
 	{
-		#define JIT_INCLUDE_RULES
-		#include "jit-rules-x86.slc"
-		#undef JIT_INCLUDE_RULES
+	#define JIT_INCLUDE_RULES
+	#include "jit-rules-x86.inc"
+	#undef JIT_INCLUDE_RULES
 
-		default:
+	default:
 		{
 			fprintf(stderr, "TODO(%x) at %s, %d\n",
-					(int)(insn->opcode), __FILE__, (int)__LINE__);
+				(int)(insn->opcode), __FILE__, (int)__LINE__);
 		}
 		break;
 	}
